﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

using System.Drawing;
using System.Drawing.Imaging;

namespace SprIpl
{
    class Program
    {
        static void Main(string[] args)
        {
            (new Program()).Start(args);
        }

        private void Usage()
        {
            Console.WriteLine("usage:");
            Console.WriteLine("SprIpl.exe command file");
            Console.WriteLine("available commands:");
            Console.WriteLine("ui = unpack IPL");
            Console.WriteLine("pi = pack IPL");
            Console.WriteLine("ui = unpack SPR");
            Console.WriteLine("pi = pack SPR");
        }

        public void Start(string[] args)
        {
            if (args.Length != 2)
            {
                Usage();
                Environment.Exit(1);
            }

            string command = args[0];
            string filename = args[1];

            if (command == "ui")
            {
                try
                {
                    UnpackIpl(filename);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                    Environment.Exit(1);
                }
            }
            else if (command == "pi")
            {
                try
                {
                    PackIpl(filename);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                    Environment.Exit(1);
                }
            }
            else if (command == "us")
            {
                try
                {
                    UnpackSpr(filename);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                    Environment.Exit(1);
                }
            }
            else if (command == "ps")
            {
                try
                {
                    throw new NotImplementedException();
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                    Environment.Exit(1);
                }
            }
            else
            {
                Usage();
                Environment.Exit(1);
            }

        }

        private void PackIpl(string filename)
        {
            FileInfo fi = new FileInfo(filename);
            Environment.CurrentDirectory = fi.DirectoryName;

            StreamReader log = new StreamReader(String.Format("{0}_log.txt", fi.Name));

            FileStream fs = fi.Open(FileMode.Create);
            BinaryWriter ipl = new BinaryWriter(fs);

            ipl.Write(Encoding.ASCII.GetBytes("1LPI"));

            int count = int.Parse(log.ReadLine());
            ipl.Write(count);

            for (int i = 0; i < count; i++)
            {
                string img_file = log.ReadLine();
                Bitmap bm = new Bitmap(img_file);

                int[] q = log.ReadLine().Split(' ').Select(x => int.Parse(x)).ToArray();
                int unk1 = q[0];
                int unk2 = q[1];
                int unk3 = q[2];

                ipl.Write(Encoding.ASCII.GetBytes("2AN\0"));
                ipl.Write(unk1);

                ipl.Write(bm.Width);
                ipl.Write(bm.Height);

                ipl.Write(unk2);
                ipl.Write(unk3);

                List<int> img_data = EncodeImage(bm);

                ipl.Write(img_data.Count);

                foreach (int d in img_data)
                    ipl.Write(d);
            }

            ipl.Close();
            fs.Close();

            log.Close();
        }

        private List<int> EncodeImage(Bitmap bm)
        {
            List<int> img = new List<int>();

            for (int y = 0; y < bm.Height; y++)
                img.AddRange(EncodeLine(bm, y));

            return img;
        }

        private List<int> EncodeLine(Bitmap bm, int y)
        {
            List<int> line = new List<int>();

            List<int> p;
            int e;

            int x = 0;
            int s = 0;

            e = ReadEmpty(bm, y, ref x);

            if (e == bm.Width)
            {
                line.Add(0);
                return line;
            }

            p = ReadPixels(bm, y, ref x);

            if (p.Count == bm.Width)
            {
                line.Add(1);
                line.AddRange(p);
                return line;
            }

            line.Add(e);
            line.Add(p.Count);
            line.AddRange(p);
            e = 0;
            p.Clear();
            s++;

            while (x < bm.Width)
            {
                e = ReadEmpty(bm, y, ref x);

                if (x >= bm.Width)
                    break;

                p = ReadPixels(bm, y, ref x);

                line.Add(e);
                line.Add(p.Count);
                line.AddRange(p);
                e = 0;
                p.Clear();
                s++;
            }

            line.Insert(0, s + 1);

            return line;
        }

        private int ReadEmpty(Bitmap bm, int y, ref int x)
        {
            int e = 0;

            while (x < bm.Width)
            {
                Color c = bm.GetPixel(x, y);
                int k = c.ToArgb();

                if (k != 0)
                    break;

                e++;
                x++;
            }

            return e;
        }

        private List<int> ReadPixels(Bitmap bm, int y, ref int x)
        {
            List<int> p = new List<int>();

            while (x < bm.Width)
            {
                Color c = bm.GetPixel(x, y);
                int k = c.ToArgb();

                if (k == 0)
                    break;

                p.Add(k);
                x++;
            }

            return p;
        }

        private void UnpackIpl(string filename)
        {
            FileInfo fi = new FileInfo(filename);
            Environment.CurrentDirectory = fi.DirectoryName;

            FileStream fs = fi.OpenRead();
            BinaryReader ipl = new BinaryReader(fs);

            StreamWriter log = new StreamWriter(String.Format("{0}_log.txt", fi.Name));

            byte[] hd1 = ipl.ReadBytes(4); // "1LPI"

            int count = ipl.ReadInt32();
            log.WriteLine(count);

            for (int i = 0; i < count; i++)
            {
                byte[] hd2 = ipl.ReadBytes(4); // "2AN\0"
                int unk1 = ipl.ReadInt32(); // 4 = has alpha-channel, 8 = has no alpha-channel

                int w = ipl.ReadInt32();
                int h = ipl.ReadInt32();

                int unk2 = ipl.ReadInt32(); // 0
                int unk3 = ipl.ReadInt32(); // 1

                int img_len = ipl.ReadInt32(); // number of 4-bytes blocks

                Bitmap bm = new Bitmap(w, h);

                for (int y = 0; y < h; y++)
                {
                    int x = 0;
                    int j = 0;

                    int t = ipl.ReadInt32();

                    //Console.WriteLine("{0:x}", t);

                    if (t == 1)
                    {
                        j = w;

                        while (j > 0)
                        {
                            Color c = Color.FromArgb(ipl.ReadInt32());
                            bm.SetPixel(x, y, c);
                            x++;
                            j--;
                        }
                    }
                    else if (t == 0)
                    {
                    }
                    else if (t > 1)
                    {
                        for (int q = 1; q < t; q++)
                        {
                            x += ipl.ReadInt32();
                            j = ipl.ReadInt32();

                            while (j > 0)
                            {
                                Color c = Color.FromArgb(ipl.ReadInt32());
                                bm.SetPixel(x, y, c);
                                x++;
                                j--;
                            }
                        }
                    }
                    else
                        throw new Exception();

                }

                string img_file = String.Format("{0}_{1}.png", fi.Name, i);

                bm.Save(img_file, ImageFormat.Png);
                log.WriteLine(img_file);
                log.WriteLine("{0} {1} {2}", unk1, unk2, unk3);
            }

            log.Close();

            ipl.Close();
            fs.Close();
        }

        private string ReadString(BinaryReader br, int n)
        {
            string s = Encoding.ASCII.GetString(br.ReadBytes(n));

            int len = s.IndexOf('\0');

            if (len >= 0)
                s = s.Substring(0, len);

            return s;
        }

        private void UnpackSpr(string filename)
        {
            FileInfo fi = new FileInfo(filename);
            Environment.CurrentDirectory = fi.DirectoryName;

            FileStream fs = fi.OpenRead();
            BinaryReader spr = new BinaryReader(fs);

            StreamWriter log = new StreamWriter(String.Format("{0}_log.txt", fi.Name));

            byte[] hd2 = spr.ReadBytes(4); // "1RPS"
            int frames_count = spr.ReadInt32();
            int movies_count = spr.ReadInt32();
            int var3 = spr.ReadInt32(); // 128
            int var4 = spr.ReadInt32(); // 32
            byte[] var5 = spr.ReadBytes(3); // 01h 1eh 03h
            byte[] var6 = spr.ReadBytes(16); // empty
            string name = ReadString(spr, 32);
            string ipl_filename = ReadString(spr, 128);

            log.WriteLine("{0} {1}", movies_count, frames_count);
            log.WriteLine("{0}", name);
            log.WriteLine("{0}", ipl_filename);

            for (int i = 0; i < movies_count; i++)
            {
                string movie_name = ReadString(spr, 32);
                short mvar1 = spr.ReadInt16(); // 0
                short first_frame = spr.ReadInt16(); // first action
                short last_frame = spr.ReadInt16(); // last action
                short mvar4 = spr.ReadInt16(); // 0
                short frames = spr.ReadInt16(); // frames count
                byte[] mvar6 = spr.ReadBytes(86); // empty

                log.WriteLine("{0}", movie_name);
                log.WriteLine("{0} {1} {2}", first_frame, last_frame, frames);
            }

            for (int i = 0; i < frames_count; i++)
            {
                short frame_id = spr.ReadInt16();
                short image = spr.ReadInt16(); // image
                short duration = spr.ReadInt16(); // duration time
                short next = spr.ReadInt16(); // next frame
                short prev = spr.ReadInt16(); // prev frame
                short x_pos2 = spr.ReadInt16(); // horizontal postion 2
                short y_pos2 = spr.ReadInt16(); // vertical position 2
                short flags = spr.ReadInt16(); // flags: 01h - horizontal flip, 02h - visible, 04h - has next, 08 - has prev
                short fvar9 = spr.ReadInt16(); // 0
                short fvar10 = spr.ReadInt16(); // 0
                short opacity = spr.ReadInt16(); // 00h-ffh
                short fvar12 = spr.ReadInt16(); // 0
                short fvar13 = spr.ReadInt16(); // 0
                short fvar14 = spr.ReadInt16(); // 0
                short x_pos = spr.ReadInt16(); // horizontal postion
                short y_pos = spr.ReadInt16(); // vertical position

                log.WriteLine("{0} {1} {2} {3} {4} {5} {6} {7} {8} {9} {10}",
                    frame_id, image, duration, next, prev, x_pos2, y_pos2, flags, opacity, x_pos, y_pos);
            }

            log.Close();

            spr.Close();
            fs.Close();
        }
    }
}
